"""Utilities for LH5 file inventory."""
from __future__ import annotations

import json
import logging
import os
import re
import string
import warnings

import h5py
import lgdo
import numpy as np
import pandas as pd
from lgdo import Array, Scalar, VectorOfVectors
from lgdo import lh5_store as lh5
from parse import parse

from . import utils

log = logging.getLogger(__name__)


class FileDB:
    """LH5 file database.

    A class containing a :class:`pandas.DataFrame` that has additional
    functions to scan the data directory, fill the dataframe's columns with
    information about each file, and read or write to disk in an LGDO format.

    The database contains the following columns:

    - file keys: the fields specified in the configuration file's
      ``file_format`` that are required to generate a file name e.g. ``run``,
      ``type``, ``timestamp`` etc.
    - ``{tier}_file``: generated file name for the tier.
    - ``{tier}_size``: size of file on disk, if applicable.
    - ``file_status``: contains a bit corresponding to whether or not a file
      for each tier exists for a given cycle e.g. If we have tiers `raw`,
      `dsp`, and `hit`, but only the `raw` file has been produced,
      ``file_status`` would be ``0b100``.
    - ``{tier}_tables``: available data streams (channels) in the tier.
    - ``{tier}_col_idx``: ``file_db.columns[{tier}_col_idx]`` will return the
      list of columns available in the tier's file.

    The database must be configured by a JSON file (or corresponding
    dictionary), which defines the data file names, paths and LH5 layout. For
    example:

    .. code-block:: json

        {
            "data_dir": "prod-ref-l200/generated/tier",
            "tier_dirs": {
                "raw": "/raw",
                "dsp": "/dsp",
                "hit": "/hit",
                "tcm": "/tcm",
                "evt": "/evt"
            },
            "file_format": {
                "raw": "/{type}/{period}/{run}/{exp}-{period}-{run}-{type}-{timestamp}-tier_raw.lh5",
                "dsp": "/{type}/{period}/{run}/{exp}-{period}-{run}-{type}-{timestamp}-tier_dsp.lh5",
                "hit": "/{type}/{period}/{run}/{exp}-{period}-{run}-{type}-{timestamp}-tier_hit.lh5",
                "evt": "/{type}/{period}/{run}/{exp}-{period}-{run}-{type}-{timestamp}-tier_evt.lh5",
                "tcm": "/{type}/{period}/{run}/{exp}-{period}-{run}-{type}-{timestamp}-tier_tcm.lh5"
            },
            "table_format": {
                "raw": "ch{ch:03d}/raw",
                "dsp": "ch{ch:03d}/dsp",
                "hit": "{ch}/hit",
                "evt": "{grp}/evt",
                "tcm": "hardware_tcm"
            },
            "tables": {
                "raw": [0, 1, 2, 4, 5, 6, 7],
                "dsp": [0, 1, 2, 4, 5, 6, 7],
                "hit": [0, 1, 2, 4, 5, 6, 7],
                "tcm": [""],
                "evt": [""]
            },
            "columns": {
                "raw": ["baseline", "waveform", "daqenergy"],
                "dsp": ["trapEftp", "AoE", "trapEmax"],
                "hit": ["trapEftp_cal", "trapEmax_cal"],
                "tcm": ["cumulative_length", "array_id", "array_idx"],
                "evt": ["lar_veto", "muon_veto", "ge_mult"]
            }
        }

    :class:`FileDB` objects can be also stored on disk and read-in at later
    times.

    Examples
    --------
    >>> from pygama.flow import FileDB
    >>> db = FileDB("./filedb_config.json")
    >>> db.scan_tables_columns()  # read in also table columns names
    >>> print(db)
    << Columns >>
    [['baseline', 'card', 'ch_orca', 'channel', 'crate', 'daqenergy', 'deadtime', 'dr_maxticks', 'dr_start_pps', 'dr_start_ticks', 'dr_stop_pps', 'dr_stop_ticks', 'eventnumber', 'fcid', 'numtraces', 'packet_id', 'runtime', 'timestamp', 'to_abs_mu_usec', 'to_dt_mu_usec', 'to_master_sec', 'to_mu_sec', 'to_mu_usec', 'to_start_sec', 'to_start_usec', 'tracelist', 'ts_maxticks', 'ts_pps', 'ts_ticks', 'waveform'], ['bl_intercept', 'bl_mean', 'bl_slope', 'bl_std', 'tail_slope', 'tail_std', 'wf_blsub'], ['array_id', 'array_idx', 'cumulative_length']]
    << DataFrame >>
       exp period   run         timestamp type  ... hit_col_idx tcm_tables tcm_col_idx evt_tables evt_col_idx
    0  l60    p01  r014  20220716T105236Z  cal  ...        None         []         [2]       None        None
    1  l60    p01  r014  20220716T104550Z  cal  ...        None         []         [2]       None        None
    >>> db.to_disk("file_db.lh5")
    """

    def __init__(self, config: str | dict | list[str], scan: bool = True) -> None:
        """
        Parameters
        ----------
        config
            dictionary or path to JSON file specifying data directories, tiers,
            and file name templates. Can also be path (or list of paths or
            regular expression) to existing LH5 file containing :class:`FileDB`
            object serialized by :meth:`.to_disk()`.
        scan
            whether the file database should scan the directory containing
            `raw` files to fill its rows with file keys.
        """
        self.df = None

        config_path = None
        if isinstance(config, str):
            config_path = config
            try:
                with open(config) as f:
                    config = json.load(f)
            # can otherwise be (wildcard of) HDF5 file(s)
            except (FileNotFoundError, json.JSONDecodeError, UnicodeDecodeError):
                self.from_disk(config)
                return

        elif isinstance(config, list):
            self.from_disk(config)
            return

        if not isinstance(config, dict):
            raise ValueError("Bad FileDB configuration value")

        self.set_config(config, config_path)

        # Set up column names
        fm = string.Formatter()
        parse_arr = np.array(list(fm.parse(self.file_format[self.tiers[0]])))
        names = list(parse_arr[:, 1])  # fields required to generate file name
        names = [n for n in names if n]  # Remove none values
        names = list(np.unique(names))
        names += [f"{tier}_file" for tier in self.tiers]  # the generated file names
        names += [f"{tier}_size" for tier in self.tiers]  # file sizes
        names += ["file_status"]  # bonus columns

        self.df = pd.DataFrame(columns=names)

        self.columns = None

        if scan:
            self.scan_files()

            # Use config columns and tables if provided
            if "columns" in self.config.keys() and "tables" in self.config.keys():
                log.debug("setting columns/tables from config")
                self.columns = list(self.config["columns"].values())
                for tier in self.tiers:
                    self.df[f"{tier}_tables"] = [self.config["tables"][tier]] * len(
                        self.df
                    )
                    self.df[f"{tier}_col_idx"] = [
                        [self.columns.index(self.config["columns"][tier])]
                        * len(self.df[f"{tier}_tables"].iloc[0])
                    ] * len(self.df)

    def set_config(self, config: dict, config_path: str = None) -> None:
        """Read in the configuration dictionary."""
        self.config = config
        self.tiers = list(self.config["tier_dirs"].keys())
        self.file_format = self.config["file_format"]
        self.table_format = self.config["table_format"]
        self.sortby = self.config.get("sortby", "timestamp")

        # expand/substitute variables in data_dir and tier_dirs
        # $_ expands to the location of the config file
        subst_vars = {}
        if config_path is not None:
            subst_vars["_"] = os.path.dirname(str(config_path))

        data_dir = lgdo.lgdo_utils.expand_path(
            self.config["data_dir"], substitute=subst_vars
        )
        self.data_dir = data_dir

        tier_dirs = self.config["tier_dirs"]
        for k, val in tier_dirs.items():
            tier_dirs[k] = lgdo.lgdo_utils.expand_vars(val, substitute=subst_vars)
        self.tier_dirs = tier_dirs

    def scan_files(self, dirs: list[str] = None) -> None:
        """Scan the directory containing files from the lowest tier and fill the dataframe.

        The lowest tier is defined as the first element of the `tiers` array.
        Only fills columns that can be populated with just these files.

        Parameters
        ----------
        dirs
            restrict search to this list of directories. Specified paths can be
            absolute, relative to `self.data_dir` or relative to the root
            directory of the lowest-tier files. If ``None``, the whole root
            lowest-tier directory is scanned. Useful to build a partial
            database.
        """
        file_keys = []
        n_files = 0
        low_tier = self.tiers[0]
        template = self.file_format[low_tier]
        root_scan_dir = os.path.join(
            self.data_dir, self.tier_dirs[low_tier].lstrip("/")
        )

        scan_dirs = dirs
        if dirs is None:
            scan_dirs = [root_scan_dir]
        elif not isinstance(dirs, list):
            scan_dirs = [dirs]

        log.info(f"scanning {scan_dirs} with template {template}")

        for scan_dir in scan_dirs:
            # some logic to guess where the scan directory is
            if not os.path.isabs(scan_dir):
                # first check if it's relative to lowest tier directory
                if os.path.isdir(os.path.join(root_scan_dir, scan_dir)):
                    scan_dir = os.path.join(root_scan_dir, scan_dir)
                # or maybe relative to the data dir?
                elif os.path.isdir(os.path.join(self.data_dir, scan_dir)):
                    scan_dir = os.path.join(self.data_dir, scan_dir)
                else:
                    scan_dir = os.path.join(os.getcwd(), scan_dir)

            log.debug(f"scanning {scan_dir}")

            for path, _, files in os.walk(scan_dir):
                log.debug(f"scanning {path}")
                n_files += len(files)

                for f in files:
                    # in some cases, we need information from the path name
                    if "/" in template:
                        f_tmp = path.replace(root_scan_dir, "") + "/" + f
                    else:
                        f_tmp = f

                    finfo = parse(template, f_tmp)
                    if finfo is not None:
                        finfo = finfo.named
                        for tier in self.tiers:
                            finfo[f"{tier}_file"] = self.file_format[tier].format(
                                **finfo
                            )

                        file_keys.append(finfo)

        if n_files == 0:
            raise FileNotFoundError(f"no {low_tier} files found")

        if len(file_keys) == 0:
            raise FileNotFoundError(f"no {low_tier} files matched pattern " + template)

        temp_df = pd.DataFrame(file_keys)

        # fill the main DataFrame
        self.df = pd.concat([self.df, temp_df])

        # convert cols to numeric dtypes where possible
        for col in self.df.columns:
            self.df[col] = pd.to_numeric(self.df[col], errors="ignore")

        # sort rows according to timestamps
        utils.inplace_sort(self.df, self.sortby)

        # set file status and sizes
        self.set_file_status()
        self.set_file_sizes()

    def set_file_status(self) -> None:
        """Add a column with a bit corresponding to whether each tier's file exists.

        For example, if we have tiers `raw`, `dsp`, and `hit`, but only the
        `raw` file has been produced, ``file_status`` would be 4 (``0b100`` in
        binary representation).
        """

        def check_status(row):
            status = 0
            for i, tier in enumerate(self.tiers):
                path_name = os.path.join(
                    self.data_dir,
                    self.tier_dirs[tier].lstrip("/"),
                    row[f"{tier}_file"].lstrip("/"),
                )
                if os.path.exists(path_name):
                    status |= 1 << len(self.tiers) - i - 1

            return status

        self.df["file_status"] = self.df.apply(check_status, axis=1)

    def set_file_sizes(self) -> None:
        """Add columns for each tier containing the corresponding file size in bytes.

        As reported by :func:`os.path.getsize`.
        """

        def get_size(row, tier):
            size = 0
            path_name = os.path.join(
                self.data_dir,
                self.tier_dirs[tier].lstrip("/"),
                row[f"{tier}_file"].lstrip("/"),
            )
            if os.path.exists(path_name):
                size = os.path.getsize(path_name)
            return size

        for tier in self.tiers:
            self.df[f"{tier}_size"] = self.df.apply(get_size, axis=1, tier=tier)

    def scan_tables_columns(
        self,
        to_file: str = None,
        override: bool = False,
        dir_files_conform: bool = False,
    ) -> list[str]:
        """Open files to read (and store) available tables (and columns therein) names.

        Adds the available table names in each tier as a column in the
        dataframe by searching for group names that match the configured
        ``table_format`` and saving the associated keyword values.

        Returns a list with each unique list of columns found in each table
        and adds a column ``{tier}_col_idx`` to the dataframe that maps to the
        column table.

        Parameters
        ----------
        to_file
            Optionally write the column table to an LH5 file (as a
            :class:`~.lgdo.vectorofvectors.VectorOfVectors`).
        override
            If the :class:`FileDB` already has a `columns` field, the scan will
            not run unless this parameter is set to ``True``.
        dir_files_conform
            if ``True``, assume that all files in a directory contain tables
            with the same columns (i.e. all file contents conform to the same
            format) and scan only the first file. Significantly reduces
            processing time.
        """
        log.info("getting table column names")

        if self.columns is not None:
            if not override:
                log.warning(
                    "LH5 tables/columns names already set, if you want to perform the scan anyway, set override=True"
                )
                return
            else:
                log.warning("overwriting existing LH5 tables/columns names")

        def update_tables_cols(row, tier: str, utc_cache: dict = None) -> pd.Series:
            fpath = os.path.join(
                self.data_dir,
                self.tier_dirs[tier].lstrip("/"),
                row[f"{tier}_file"].lstrip("/"),
            )
            this_dir = fpath[: fpath.rfind("/")]
            if utc_cache is not None and this_dir in utc_cache:
                return utc_cache[this_dir]

            log.debug(f"reading column names for tier '{tier}' from {fpath}")

            if os.path.exists(fpath):
                f = h5py.File(fpath)
            else:
                log.debug(f"{fpath} doesn't exist")
                return pd.Series({f"{tier}_tables": None, f"{tier}_col_idx": None})

            # Get tables in each tier
            tier_tables = []
            template = self.table_format[tier]
            if template[-1] == "/":
                template = template[:-1]

            braces = list(re.finditer("{|}", template))

            if len(braces) > 2:
                raise ValueError("tables can only have one identifier")
            if len(braces) % 2 != 0:
                raise ValueError("braces mismatch in table format")
            if len(braces) == 0:
                tier_tables.append(0)
            else:
                wildcard = (
                    template[: braces[0].span()[0]]
                    + "*"
                    + template[braces[1].span()[1] :]
                )

                # TODO this call here is really expensive!
                groups = lh5.ls(f, wildcard)
                if len(groups) > 0 and parse(template, groups[0]) is None:
                    log.warning(f"groups in {fpath} don't match template")
                else:
                    tier_tables = [
                        list(parse(template, g).named.values())[0] for g in groups
                    ]

            # Get columns
            col_idx = []
            template = self.table_format[tier]
            fm = string.Formatter()

            for tb in tier_tables:
                parse_arr = np.array(list(fm.parse(template)))
                names = list(parse_arr[:, 1])
                if len(names) > 0:
                    keyword = names[0]
                    args = {keyword: tb}
                    table_name = template.format(**args)
                else:
                    table_name = template

                try:
                    col = lh5.ls(f[table_name])
                except KeyError:
                    log.warning(f"cannot find '{table_name}' in {fpath}")
                    continue
                if col not in columns:
                    columns.append(col)
                    col_idx.append(len(columns) - 1)
                else:
                    col_idx.append(columns.index(col))

            series = pd.Series(
                {f"{tier}_tables": tier_tables, f"{tier}_col_idx": col_idx}
            )
            if utc_cache is not None:
                utc_cache[this_dir] = series
            return series

        columns = []

        # set up a cache to provide a fast option if all files in each directory
        # are expected to all have the same cols
        utc_cache = None
        if dir_files_conform:
            utc_cache = {}

        for tier in self.tiers:
            self.df[[f"{tier}_tables", f"{tier}_col_idx"]] = self.df.apply(
                update_tables_cols, axis=1, tier=tier, utc_cache=utc_cache
            )

        self.columns = columns

        if to_file is not None:
            log.debug(f"writing column names to '{to_file}'")
            flattened = []
            length = []
            for i, col in enumerate(columns):
                if i == 0:
                    length.append(len(col))
                else:
                    length.append(length[i - 1] + len(col))
                for c in col:
                    flattened.append(c)
            columns_vov = VectorOfVectors(
                flattened_data=flattened, cumulative_length=length
            )
            sto = lh5.LH5Store()
            sto.write_object(columns_vov, "unique_columns", to_file)

        return self.columns

    def from_disk(self, path: str | list[str]) -> None:
        """Read FileDBs from disk.

        Overrides the dataframe, configuration dictionary and columns with the
        information from a file created by :meth:`to_disk`.

        Parameters
        ----------
        path
            file or file pattern (or list of the latter).
        """
        log.debug(f"reading FileDB from disk at {path}")

        if not isinstance(path, list):
            path = [path]

        # expand wildcards
        paths = []
        for p in path:
            paths += lgdo.lgdo_utils.expand_path(p, list=True)

        if not paths:
            raise FileNotFoundError(path)

        sto = lh5.LH5Store()
        # objects/accumulators that will be used to configure the FileDB at the end
        _cfg = None
        _df = None
        _columns = None

        # function needed later in the loop
        def _replace_idx(row, trans, tier):
            col = row[f"{tier}_col_idx"]
            if col is None:
                return None

            col = np.array(col)
            new_col = np.copy(col)

            for idx, new_idx in trans.items():
                new_col[np.where(col == idx)] = new_idx

            return new_col.tolist()

        # loop over the files
        for p in paths:
            cfg, _ = sto.read_object("config", p)
            cfg = json.loads(cfg.value.decode())

            # make sure configurations are all the same
            if _cfg is None:
                _cfg = cfg
            elif cfg != _cfg:
                raise RuntimeError(
                    "cannot merge FileDBs created with different configuration files"
                )

            # read in unique columns
            vov, _ = sto.read_object("columns", p)
            # Convert back from VoV of UTF-8 bytestrings to a list of lists of strings
            columns = [[v.decode("utf-8") for v in ov] for ov in list(vov)]

            # read in dataframe
            df = pd.read_hdf(p, key="dataframe")

            # first iteration
            if _columns is None:
                _columns = columns
                _df = df
                continue

            elif _columns != columns:
                log.debug("found inconsistent FileDB, trying to merge")
                # if columns are not the same, need to merge the two dataframes
                # in the right way. loop over new columns
                idx_trans = {}
                for idx, cols in enumerate(columns):
                    new_idx = None

                    # the columns might be a new entry...
                    if cols not in _columns:
                        # add the new column at the end and save its index
                        _columns += [cols]
                        new_idx = len(_columns) - 1
                    # ...or just located (at a different index?) in the
                    # existing column list
                    else:
                        new_idx = _columns.index(cols)

                    idx_trans[idx] = new_idx

                # now go through the new dataframe and update the old index
                # everywhere in the {tier}_col_idx columns
                for tier in list(_cfg["tier_dirs"].keys()):
                    df[f"{tier}_col_idx"] = df.apply(
                        _replace_idx,
                        args=(idx_trans, tier),
                        axis=1,
                    )

            # now we can safely concat the dataframes
            _df = pd.concat([_df, df], ignore_index=True, copy=False)

        self.set_config(_cfg)
        self.df = _df
        self.columns = _columns

        utils.inplace_sort(self.df, self.sortby)

    def to_disk(self, filename: str, wo_mode="write_safe") -> None:
        """Serializes database to disk.

        Parameters
        -----------
        filename
            output LH5 file name.
        wo_mode
            passed to :meth:`~.lgdo.lh5_store.write_object`.
        """
        log.debug(f"writing database to {filename}")

        sto = lh5.LH5Store()
        sto.write_object(
            Scalar(json.dumps(self.config)), "config", filename, wo_mode=wo_mode
        )

        if wo_mode in ["write_safe", "w", "overwrite_file", "of"]:
            wo_mode = "a"

        if self.columns is not None:
            flat = []
            cum_l = [0]
            for i in range(len(self.columns)):
                flat += self.columns[i]
                cum_l.append(cum_l[i] + len(self.columns[i]))
            cum_l = cum_l[1:]
            # Must use type 'S' to play nice with HDF
            col_vov = VectorOfVectors(
                flattened_data=Array(nda=np.array(flat).astype("S")),
                cumulative_length=Array(nda=np.array(cum_l)),
            )
            sto.write_object(col_vov, "columns", filename, wo_mode=wo_mode)

        # FIXME: to_hdf() throws this:
        #
        #     pandas.errors.PerformanceWarning: your performance may suffer as
        #     PyTables will pickle object types that it cannot map directly to c-types
        #
        # not sure how to fix this so we ignore the warning for the moment
        warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)
        self.df.to_hdf(filename, key="dataframe", format="fixed", mode="r+")

    def scan_daq_files(self, daq_dir: str, daq_template: str) -> None:
        """
        Does the exact same thing as :meth:`.scan_files` but with extra
        configuration arguments for a DAQ directory and template instead of
        using the lowest tier.
        """
        file_keys = []
        n_files = 0

        for path, _folders, files in os.walk(daq_dir):
            n_files += len(files)

            for f in files:
                # in some cases, we need information from the path name
                if "/" in daq_template:
                    f_tmp = path.replace(daq_dir, "") + "/" + f
                else:
                    f_tmp = f

                finfo = parse(daq_template, f_tmp)
                if finfo is not None:
                    finfo = finfo.named
                    file_keys.append(finfo)
                for tier in self.tiers:
                    finfo[f"{tier}_file"] = self.file_format[tier].format(**finfo)

        if n_files == 0:
            raise FileNotFoundError("No DAQ files found")

        if len(file_keys) == 0:
            raise FileNotFoundError("No DAQ files matched pattern ", daq_template)

        temp_df = pd.DataFrame(file_keys)

        # fill the main DataFrame
        self.df = pd.concat([self.df, temp_df])

        # convert cols to numeric dtypes where possible
        for col in self.df.columns:
            self.df[col] = pd.to_numeric(self.df[col], errors="ignore")

    def get_table_name(self, tier: str, tb: str) -> str:
        """Get the table name for a tier given its table identifier.

        Parameters
        ----------
        tier
            specify the tier whose table format will be used.
        tb
            the table identifier that will be passed to the table format.

        Returns
        -------
        table_name
            the name of the table in `tier` with table identifier `tb`
        """
        template = self.table_format[tier]
        fm = string.Formatter()
        parse_arr = np.array(list(fm.parse(template)))
        names = list(parse_arr[:, 1])
        if len(names) > 0:
            keyword = names[0]
            args = {keyword: tb}
            table_name = template.format(**args)
        else:
            table_name = template
        return table_name

    def get_table_columns(
        self, table: str | int, tier: str, ifile: int = 0
    ) -> list[str]:
        """Return list of columns in table `table`, tier `tier`.

        Assumes that the table contents do not change across data files. If
        desired, `ifile` (default is 0) can be used to select a different file.
        """
        tables = self.df.iloc[ifile][f"{tier}_tables"]
        if tables is None:
            return []

        table_idx = tables.index(table)
        col_idx = self.df.iloc[ifile][f"{tier}_col_idx"][table_idx]
        return self.columns[col_idx]

    def __repr__(self) -> str:
        string = f"FileDB(data_dir={self.data_dir}, tiers={self.tier_dirs}, "

        if self.df is not None:
            string += "df=DataFrame(...), "
        else:
            string += "df=None, "

        if self.columns is not None:
            string += "columns=[...]"
        else:
            string += "columns=None"

        return string + ")"



from warnings import warn

from daq2lh5 import *  # noqa: F403, F401

warn(
    "pygama.raw has moved to its own package, legend-daq2lh5. "
    "Please replace 'import pygama.raw' with 'import daq2lh5'. "
    "pygama.raw will be removed in a future release.",
    DeprecationWarning,
    stacklevel=2,
)



"""
This module provides a routine for running the energy calibration on Th data
"""

from __future__ import annotations

import json
import logging
import math
import os
import pathlib
from datetime import datetime

import matplotlib as mpl
from scipy.stats import binned_statistic

mpl.use("agg")
import lgdo.lh5_store as lh5
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats
from iminuit import Minuit, cost
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import LogNorm
from scipy.optimize import curve_fit

import pygama.math.histogram as pgh
import pygama.math.peak_fitting as pgf
import pygama.pargen.cuts as cts
import pygama.pargen.energy_cal as cal
from pygama.pargen.utils import load_data, return_nans

log = logging.getLogger(__name__)


def fwhm_slope(x: np.array, m0: float, m1: float, m2: float = None) -> np.array:
    """
    Fit the energy resolution curve
    """
    if m2 is None:
        return np.sqrt(m0 + m1 * x)
    else:
        return np.sqrt(m0 + m1 * x + m2 * x**2)


def apply_cuts(
    data: pd.DataFrame,
    hit_dict,
    cut_parameters=None,
    final_cut_field: str = "is_valid_cal",
    pulser_field="is_pulser",
):
    if cut_parameters is not None:
        cut_dict = cts.generate_cuts(data.query(f"(~{pulser_field})"), cut_parameters)
        hit_dict.update(
            cts.cut_dict_to_hit_dict(cut_dict, final_cut_field=final_cut_field)
        )
        mask = cts.get_cut_indexes(data, cut_dict)

        data[final_cut_field] = mask

    else:
        data[final_cut_field] = np.ones(len(data), dtype=bool)

    events_pqc = len(data.query(f"{final_cut_field}&(~{pulser_field})"))
    log.debug(f"{events_pqc} events valid for calibration")

    return data, hit_dict


def gen_pars_dict(pars, deg, energy_param):
    if deg == 1:
        out_dict = {
            "expression": f"a*{energy_param}+b",
            "parameters": {"a": pars[0], "b": pars[1]},
        }
    elif deg == 0:
        out_dict = {
            "expression": f"a*{energy_param}",
            "parameters": {"a": pars[0]},
        }
    elif deg == 2:
        out_dict = {
            "expression": f"a*{energy_param}**2 +b*{energy_param}+c",
            "parameters": {"a": pars[0], "b": pars[1], "c": pars[2]},
        }
    else:
        out_dict = {}
        log.error(f"hit_dict not implemented for deg = {deg}")

    return out_dict


class fwhm_linear:
    def func(x, a, b):
        return np.sqrt(a + b * x)

    def string_func(input_param):
        return f"(a+b*{input_param})**(0.5)"

    def guess(xs, ys, y_errs):
        return [np.nanmin(ys), 10**-3]

    def bounds():
        return [(0, None), (0, None)]


class fwhm_quadratic:
    def func(x, a, b, c):
        return np.sqrt(a + b * x + c * x**2)

    def string_func(input_param):
        return f"(a+b*{input_param}+c*{input_param}**2)**(0.5)"

    def guess(xs, ys, y_errs):
        return [np.nanmin(ys), 10**-3, 10**-5]

    def bounds():
        return [(0, None), (0, None), (0, None)]


class calibrate_parameter:
    glines = [
        # 238.632,
        583.191,
        727.330,
        860.564,
        1592.53,
        1620.50,
        2103.53,
        2614.50,
    ]  # gamma lines used for calibration
    range_keV = [
        # (8, 8),
        (20, 20),
        (30, 30),
        (30, 30),
        (40, 20),
        (20, 40),
        (40, 40),
        (60, 60),
    ]  # side bands width
    funcs = [
        # pgf.extended_gauss_step_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_radford_pdf,
    ]
    gof_funcs = [
        # pgf.gauss_step_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
        pgf.radford_pdf,
    ]

    def __init__(
        self,
        energy_param,
        selection_string="",
        plot_options: dict = None,
        guess_keV: float | None = None,
        threshold: int = 0,
        p_val: float = 0,
        n_events: int = None,
        simplex: bool = True,
        deg: int = 1,
        cal_energy_param: str = None,
        tail_weight=100,
    ):
        self.energy_param = energy_param
        if cal_energy_param is None:
            self.cal_energy_param = f"{self.energy_param}_cal"
        else:
            self.cal_energy_param = cal_energy_param
        self.selection_string = selection_string
        self.guess_keV = guess_keV
        self.threshold = threshold
        self.p_val = p_val
        self.n_events = n_events
        self.deg = deg
        self.plot_options = plot_options
        self.simplex = simplex
        self.tail_weight = tail_weight

        self.output_dict = {}
        self.hit_dict = {}

    def fit_energy_res(self):
        fitted_peaks = self.results["fitted_keV"]
        fwhms = self.results["pk_fwhms"][:, 0]
        dfwhms = self.results["pk_fwhms"][:, 1]

        #####
        # Remove the Tl SEP and DEP from calibration if found
        fwhm_peaks = np.array([], dtype=np.float32)
        all_peaks = np.array([], dtype=np.float32)
        indexes = []
        for i, peak in enumerate(fitted_peaks):
            all_peaks = np.append(all_peaks, peak)
            if peak == 2103.53:
                log.info(f"Tl SEP found at index {i}")
                indexes.append(i)
                continue
            elif peak == 1592.53:
                log.info(f"Tl DEP found at index {i}")
                indexes.append(i)
                continue
            elif peak == 511.0:
                log.info(f"e annhilation found at index {i}")
                indexes.append(i)
                continue
            elif np.isnan(dfwhms[i]):
                log.info(f"{peak} failed")
                indexes.append(i)
                continue
            else:
                fwhm_peaks = np.append(fwhm_peaks, peak)
        fit_fwhms = np.delete(fwhms, [indexes])
        fit_dfwhms = np.delete(dfwhms, [indexes])
        #####
        for i, peak in enumerate(fwhm_peaks):
            log.info(
                f"FWHM of {peak} keV peak is: {fit_fwhms[i]:1.2f} +- {fit_dfwhms[i]:1.2f} keV"
            )
        try:
            if 2614.50 not in fwhm_peaks:
                raise RuntimeError

            c_lin = cost.LeastSquares(
                fwhm_peaks, fit_fwhms, fit_dfwhms, fwhm_linear.func
            )
            c_lin.loss = "soft_l1"
            m_lin = Minuit(c_lin, *fwhm_linear.guess(fwhm_peaks, fit_fwhms, fit_dfwhms))
            m_lin.limits = fwhm_linear.bounds()
            m_lin.simplex()
            m_lin.migrad()
            m_lin.hesse()

            rng = np.random.default_rng(1)
            pars_b = rng.multivariate_normal(m_lin.values, m_lin.covariance, size=1000)
            fits = np.array([fwhm_linear.func(fwhm_peaks, *par_b) for par_b in pars_b])
            qbb_vals = np.array([fwhm_linear.func(2039.0, *par_b) for par_b in pars_b])
            qbb_err = np.nanstd(qbb_vals)
            predicted_fwhms = fwhm_linear.func(fwhm_peaks, *m_lin.values)
            fit_qbb = fwhm_linear.func(2039.0, *m_lin.values)

            p_val = scipy.stats.chi2.sf(m_lin.fval, len(fwhm_peaks) - len(m_lin.values))

            self.fwhm_fit_linear = {
                "function": fwhm_linear.__name__,
                "module": fwhm_linear.__module__,
                "expression": fwhm_linear.string_func("x"),
                "Qbb_fwhm_in_keV": fit_qbb,
                "Qbb_fwhm_err_in_keV": qbb_err,
                "parameters": m_lin.values,
                "uncertainties": m_lin.errors,
                "cov": m_lin.covariance,
                "csqr": (m_lin.fval, len(fwhm_peaks) - len(m_lin.values)),
                "p_val": p_val,
            }

            log.info(f'FWHM linear fit: {self.fwhm_fit_linear["parameters"].to_dict()}')
            log.info(f"FWHM fit values:")
            log.info(f"\t   Energy   | FWHM (keV)  | Predicted (keV)")
            for i, (peak, fwhm, fwhme) in enumerate(
                zip(fwhm_peaks, fit_fwhms, fit_dfwhms)
            ):
                log.info(
                    f"\t{i}".ljust(4)
                    + str(peak).ljust(9)
                    + f"| {fwhm:.2f}+-{fwhme:.2f}  ".ljust(5)
                    + f"| {fwhm_linear.func(peak, *self.fwhm_fit_linear['parameters']):.2f}".ljust(
                        5
                    )
                )

            log.info(
                f"FWHM energy resolution at Qbb (linear fit): {fit_qbb:1.2f} +- {qbb_err:1.2f} keV"
            )
        except RuntimeError:
            log.error(f"FWHM linear fit failed for {self.energy_param}")
            pars, errs, cov = return_nans(fwhm_linear.func)
            self.fwhm_fit_linear = {
                "function": fwhm_linear.__name__,
                "module": fwhm_linear.__module__,
                "expression": fwhm_linear.string_func("x"),
                "Qbb_fwhm_in_keV": np.nan,
                "Qbb_fwhm_err_in_keV": np.nan,
                "parameters": pars,
                "uncertainties": errs,
                "cov": cov,
                "csqr": (np.nan, np.nan),
                "p_val": 0,
            }
            log.error("FWHM linear fit failed to converge")
        try:
            if 2614.50 not in fwhm_peaks:
                raise RuntimeError
            c_quad = cost.LeastSquares(
                fwhm_peaks, fit_fwhms, fit_dfwhms, fwhm_quadratic.func
            )
            c_quad.loss = "soft_l1"
            m_quad = Minuit(
                c_quad, *fwhm_quadratic.guess(fwhm_peaks, fit_fwhms, fit_dfwhms)
            )
            m_quad.limits = fwhm_quadratic.bounds()
            m_quad.simplex()
            m_quad.migrad()
            m_quad.hesse()

            rng = np.random.default_rng(1)
            pars_b = rng.multivariate_normal(
                m_quad.values, m_quad.covariance, size=1000
            )
            fits = np.array(
                [fwhm_quadratic.func(fwhm_peaks, *par_b) for par_b in pars_b]
            )
            qbb_vals = np.array(
                [fwhm_quadratic.func(2039.0, *par_b) for par_b in pars_b]
            )
            qbb_err = np.nanstd(qbb_vals)
            predicted_fwhms = fwhm_quadratic.func(fwhm_peaks, *m_quad.values)
            fit_qbb = fwhm_quadratic.func(2039.0, *m_quad.values)

            p_val = scipy.stats.chi2.sf(
                m_quad.fval, len(fwhm_peaks) - len(m_quad.values)
            )

            self.fwhm_fit_quadratic = {
                "function": fwhm_quadratic.__name__,
                "module": fwhm_quadratic.__module__,
                "expression": fwhm_quadratic.string_func("x"),
                "Qbb_fwhm_in_keV": fit_qbb,
                "Qbb_fwhm_err_in_keV": qbb_err,
                "parameters": m_quad.values,
                "uncertainties": m_quad.errors,
                "cov": m_quad.covariance,
                "csqr": (m_quad.fval, len(fwhm_peaks) - len(m_quad.values)),
                "p_val": p_val,
            }
            log.info(
                f'FWHM quadratic fit: {self.fwhm_fit_quadratic["parameters"].to_dict()}'
            )
            log.info(
                f"FWHM energy resolution at Qbb (quadratic fit): {fit_qbb:1.2f} +- {qbb_err:1.2f} keV"
            )
        except RuntimeError:
            log.error(f"FWHM quadratic fit failed for {self.energy_param}")
            pars, errs, cov = return_nans(fwhm_quadratic.func)
            self.fwhm_fit_quadratic = {
                "function": fwhm_quadratic.__name__,
                "module": fwhm_quadratic.__module__,
                "expression": fwhm_quadratic.string_func("x"),
                "Qbb_fwhm_in_keV": np.nan,
                "Qbb_fwhm_err_in_keV": np.nan,
                "parameters": pars,
                "uncertainties": errs,
                "cov": cov,
                "csqr": (np.nan, np.nan),
                "p_val": 0,
            }
            log.error("FWHM quadratic fit failed to converge")

    def gen_pars_dict(self):
        if self.deg == 1:
            out_dict = {
                "expression": f"a*{self.energy_param}+b",
                "parameters": {"a": self.pars[0], "b": self.pars[1]},
            }
        elif self.deg == 0:
            out_dict = {
                "expression": f"a*{self.energy_param}",
                "parameters": {"a": self.pars[0]},
            }
        elif self.deg == 2:
            out_dict = {
                "expression": f"a*{self.energy_param}**2 +b*{self.energy_param}+c",
                "parameters": {"a": self.pars[0], "b": self.pars[1], "c": self.pars[2]},
            }
        else:
            out_dict = {}
            log.warning(f"hit_dict not implemented for deg = {self.deg}")

        return out_dict

    def get_results_dict(self, data):
        if np.isnan(self.pars).all():
            return {}
        else:
            fwhm_linear = self.fwhm_fit_linear.copy()
            fwhm_linear["parameters"] = fwhm_linear["parameters"].to_dict()
            fwhm_linear["uncertainties"] = fwhm_linear["uncertainties"].to_dict()
            fwhm_linear["cov"] = fwhm_linear["cov"].tolist()
            fwhm_quad = self.fwhm_fit_quadratic.copy()
            fwhm_quad["parameters"] = fwhm_quad["parameters"].to_dict()
            fwhm_quad["uncertainties"] = fwhm_quad["uncertainties"].to_dict()
            fwhm_quad["cov"] = fwhm_quad["cov"].tolist()

            pk_dict = {
                Ei: {
                    "function": func_i.__name__,
                    "module": func_i.__module__,
                    "parameters_in_ADC": parsi.to_dict(),
                    "uncertainties_in_ADC": errorsi.to_dict(),
                    "p_val": pvali,
                    "fwhm_in_keV": list(fwhmi),
                }
                for i, (Ei, parsi, errorsi, pvali, fwhmi, func_i) in enumerate(
                    zip(
                        self.results["fitted_keV"],
                        self.results["pk_pars"][self.results["pk_validities"]],
                        self.results["pk_errors"][self.results["pk_validities"]],
                        self.results["pk_pvals"][self.results["pk_validities"]],
                        self.results["pk_fwhms"],
                        self.funcs,
                    )
                )
            }

            return {
                "total_fep": len(
                    data.query(
                        f"{self.cal_energy_param}>2604&{self.cal_energy_param}<2624"
                    )
                ),
                "total_dep": len(
                    data.query(
                        f"{self.cal_energy_param}>1587&{self.cal_energy_param}<1597"
                    )
                ),
                "pass_fep": len(
                    data.query(
                        f"{self.cal_energy_param}>2604&{self.cal_energy_param}<2624&{self.selection_string}"
                    )
                ),
                "pass_dep": len(
                    data.query(
                        f"{self.cal_energy_param}>1587&{self.cal_energy_param}<1597&{self.selection_string}"
                    )
                ),
                "eres_linear": fwhm_linear,
                "eres_quadratic": fwhm_quad,
                "fitted_peaks": self.results["fitted_keV"].tolist(),
                "pk_fits": pk_dict,
            }

    def calibrate_parameter(self, data):
        kev_ranges = self.range_keV.copy()
        if self.guess_keV is None:
            self.guess_keV = 2620 / np.nanpercentile(
                data.query(
                    f"{self.selection_string} & {self.energy_param}>{self.threshold}"
                )[self.energy_param],
                99,
            )

        log.debug(f"Find peaks and compute calibration curve for {self.energy_param}")
        log.debug(f"Guess is {self.guess_keV:.3f}")

        try:
            self.pars, self.cov, self.results = cal.hpge_E_calibration(
                data.query(self.selection_string)[self.energy_param],
                self.glines,
                self.guess_keV,
                deg=self.deg,
                range_keV=kev_ranges,
                funcs=self.funcs,
                gof_funcs=self.gof_funcs,
                n_events=self.n_events,
                allowed_p_val=self.p_val,
                simplex=self.simplex,
                tail_weight=self.tail_weight,
                verbose=False,
            )
            pk_pars = self.results["pk_pars"]
            found_peaks = self.results["got_peaks_locs"]
            fitted_peaks = self.results["fitted_keV"]
            fitted_funcs = self.results["pk_funcs"]
            if self.pars is None:
                raise ValueError

            for i, peak in enumerate(self.results["got_peaks_keV"]):
                idx = np.where(peak == self.glines)[0][0]
                self.funcs[idx] = fitted_funcs[i]
                if fitted_funcs[i] == pgf.extended_radford_pdf:
                    self.gof_funcs[idx] = pgf.radford_pdf
                else:
                    self.gof_funcs[idx] = pgf.gauss_step_pdf
        except:
            found_peaks = np.array([])
            fitted_peaks = np.array([])
            fitted_funcs = np.array([])
        if len(fitted_peaks) != len(self.glines):
            for i, peak in enumerate(self.glines):
                if peak not in fitted_peaks:
                    kev_ranges[i] = (kev_ranges[i][0] - 5, kev_ranges[i][1] - 5)
            for i, peak in enumerate(fitted_peaks):
                try:
                    if (
                        self.results["pk_fwhms"][:, 1][i]
                        / self.results["pk_fwhms"][:, 0][i]
                        > 0.05
                    ):
                        index = np.where(self.glines == peak)[0][0]
                        kev_ranges[index] = (
                            kev_ranges[index][0] - 5,
                            kev_ranges[index][1] - 5,
                        )
                except:
                    pass
            try:
                self.pars, self.cov, self.results = cal.hpge_E_calibration(
                    data.query(self.selection_string)[self.energy_param],
                    self.glines,
                    self.guess_keV,
                    deg=self.deg,
                    range_keV=kev_ranges,
                    funcs=self.funcs,
                    gof_funcs=self.gof_funcs,
                    n_events=self.n_events,
                    allowed_p_val=self.p_val,
                    simplex=self.simplex,
                    tail_weight=self.tail_weight,
                    verbose=False,
                )
                fitted_peaks = self.results["fitted_keV"]
                fitted_funcs = self.results["pk_funcs"]

                log.debug("Calibrated found")
                log.info(f"Calibration pars are {self.pars}")

                for i, peak in enumerate(self.results["got_peaks_keV"]):
                    idx = np.where(peak == self.glines)[0][0]
                    self.funcs[idx] = fitted_funcs[i]
                    if fitted_funcs[i] == pgf.extended_radford_pdf:
                        self.gof_funcs[idx] = pgf.radford_pdf
                    else:
                        self.gof_funcs[idx] = pgf.gauss_step_pdf
                if self.pars is None:
                    raise ValueError

            except:
                self.pars = np.full(self.deg + 1, np.nan)
                self.results = None

                log.error(f"Calibration failed completely for {self.energy_param}")
        else:
            log.debug("Calibrated found")
            log.info(f"Calibration pars are {self.pars}")
        if ~np.isnan(self.pars).all():
            self.fit_energy_res()
        self.hit_dict[self.cal_energy_param] = self.gen_pars_dict()
        data[f"{self.energy_param}_cal"] = pgf.poly(data[self.energy_param], self.pars)

    def fill_plot_dict(self, data, plot_dict={}):
        for key, item in self.plot_options.items():
            if item["options"] is not None:
                plot_dict[key] = item["function"](self, data, **item["options"])
            else:
                plot_dict[key] = item["function"](self, data)
        return plot_dict


class high_stats_fitting(calibrate_parameter):
    glines = [
        238.632,
        511,
        583.191,
        727.330,
        763,
        785,
        860.564,
        893,
        1079,
        1513,
        1592.53,
        1620.50,
        2103.53,
        2614.50,
        3125,
        3198,
        3474,
    ]  # gamma lines used for calibration
    range_keV = [
        (10, 10),
        (30, 30),
        (30, 30),
        (30, 30),
        (30, 15),
        (15, 30),
        (30, 25),
        (25, 30),
        (30, 30),
        (30, 30),
        (30, 20),
        (20, 30),
        (30, 30),
        (30, 30),
        (30, 30),
        (30, 30),
        (30, 30),
    ]  # side bands width
    binning = [
        0.02,
        0.02,
        0.02,
        0.02,
        0.2,
        0.2,
        0.02,
        0.2,
        0.2,
        0.2,
        0.1,
        0.1,
        0.1,
        0.02,
        0.2,
        0.2,
        0.2,
    ]
    funcs = [
        pgf.extended_gauss_step_pdf,  # probably should be gauss on exp
        pgf.extended_gauss_step_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_radford_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_gauss_step_pdf,
        pgf.extended_gauss_step_pdf,
    ]
    gof_funcs = [
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
        pgf.radford_pdf,
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.radford_pdf,
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
        pgf.gauss_step_pdf,
    ]

    def __init__(
        self,
        energy_param,
        selection_string,
        threshold,
        p_val,
        plot_options={},
        simplex=False,
        tail_weight=20,
    ):
        self.energy_param = energy_param
        self.cal_energy_param = energy_param
        self.selection_string = selection_string
        self.threshold = threshold
        self.p_val = p_val
        self.plot_options = plot_options
        self.simplex = simplex
        self.results = {}
        self.plot_dict = {}
        self.n_events = None
        self.output_dict = {}
        self.pars = [1, 0]
        self.tail_weight = tail_weight

    def get_results_dict(self, data):
        if self.results:
            fwhm_linear = self.fwhm_fit_linear.copy()
            fwhm_linear["parameters"] = fwhm_linear["parameters"].to_dict()
            fwhm_linear["uncertainties"] = fwhm_linear["uncertainties"].to_dict()
            fwhm_linear["cov"] = fwhm_linear["cov"].tolist()
            fwhm_quad = self.fwhm_fit_quadratic.copy()
            fwhm_quad["parameters"] = fwhm_quad["parameters"].to_dict()
            fwhm_quad["uncertainties"] = fwhm_quad["uncertainties"].to_dict()
            fwhm_quad["cov"] = fwhm_quad["cov"].tolist()

            pk_dict = {
                Ei: {
                    "function": func_i.__name__,
                    "module": func_i.__module__,
                    "parameters_in_keV": parsi.to_dict(),
                    "uncertainties_in_keV": errorsi.to_dict(),
                    "p_val": pvali,
                    "fwhm_in_keV": list(fwhmi),
                }
                for i, (Ei, parsi, errorsi, pvali, fwhmi, func_i) in enumerate(
                    zip(
                        self.results["fitted_keV"],
                        self.results["pk_pars"][self.results["pk_validities"]],
                        self.results["pk_errors"][self.results["pk_validities"]],
                        self.results["pk_pvals"][self.results["pk_validities"]],
                        self.results["pk_fwhms"],
                        self.funcs,
                    )
                )
            }

            return {
                "eres_linear": fwhm_linear,
                "eres_quadratic": fwhm_quad,
                "fitted_peaks": self.results["fitted_keV"].tolist(),
                "pk_fits": pk_dict,
            }
        else:
            return {}

    def fit_peaks(self, data):
        log.debug(f"Fitting {self.energy_param}")
        try:
            n_bins = [
                int((self.range_keV[i][1] + self.range_keV[i][0]) / self.binning[i])
                for i in range(len(self.glines))
            ]
            (
                pk_pars,
                pk_errors,
                pk_covs,
                pk_binws,
                pk_ranges,
                pk_pvals,
                valid_pks,
                pk_funcs,
            ) = cal.hpge_fit_E_peaks(
                data.query(self.selection_string)[self.energy_param],
                self.glines,
                self.range_keV,
                n_bins=n_bins,
                funcs=self.funcs,
                method="unbinned",
                gof_funcs=self.gof_funcs,
                n_events=None,
                allowed_p_val=self.p_val,
                tail_weight=20,
            )
            for idx, peak in enumerate(self.glines):
                self.funcs[idx] = pk_funcs[idx]
                if pk_funcs[idx] == pgf.extended_radford_pdf:
                    self.gof_funcs[idx] = pgf.radford_pdf
                else:
                    self.gof_funcs[idx] = pgf.gauss_step_pdf

            self.results["got_peaks_keV"] = self.glines
            self.results["pk_pars"] = pk_pars
            self.results["pk_errors"] = pk_errors
            self.results["pk_covs"] = pk_covs
            self.results["pk_binws"] = pk_binws
            self.results["pk_ranges"] = pk_ranges
            self.results["pk_pvals"] = pk_pvals

            for i, pk in enumerate(self.results["got_peaks_keV"]):
                try:
                    if self.results["pk_pars"][i]["n_sig"] < 10:
                        valid_pks[i] = False
                    elif (
                        2 * self.results["pk_errors"][i]["n_sig"]
                        > self.results["pk_pars"][i]["n_sig"]
                    ):
                        valid_pks[i] = False
                except:
                    pass

            self.results["pk_validities"] = valid_pks

            # Drop failed fits
            fitted_peaks_keV = self.results["fitted_keV"] = np.asarray(self.glines)[
                valid_pks
            ]
            pk_pars = np.asarray(pk_pars, dtype=object)[valid_pks]  # ragged
            pk_errors = np.asarray(pk_errors, dtype=object)[valid_pks]
            pk_covs = np.asarray(pk_covs, dtype=object)[valid_pks]
            pk_binws = np.asarray(pk_binws)[valid_pks]
            pk_ranges = np.asarray(pk_ranges)[valid_pks]
            pk_pvals = np.asarray(pk_pvals)[valid_pks]
            pk_funcs = np.asarray(pk_funcs)[valid_pks]

            log.info(f"{len(np.where(valid_pks)[0])} peaks fitted:")
            for i, (Ei, parsi, errorsi, covsi, func_i) in enumerate(
                zip(fitted_peaks_keV, pk_pars, pk_errors, pk_covs, pk_funcs)
            ):
                varnames = func_i.__code__.co_varnames[1 : len(pk_pars[-1]) + 1]
                parsi = np.asarray(parsi, dtype=float)
                errorsi = np.asarray(errorsi, dtype=float)
                covsi = np.asarray(covsi, dtype=float)
                # parsigsi = np.sqrt(covsi.diagonal())
                log.info(f"\tEnergy: {str(Ei)}")
                log.info(f"\t\tParameter  |    Value +/- Sigma  ")
                for vari, pari, errorsi in zip(varnames, parsi, errorsi):
                    log.info(
                        f'\t\t{str(vari).ljust(10)} | {("%4.2f" % pari).rjust(8)} +/- {("%4.2f" % errorsi).ljust(8)}'
                    )

            cal_fwhms = [
                pgf.get_fwhm_func(func_i, pars_i, cov=covs_i)
                for func_i, pars_i, covs_i in zip(pk_funcs, pk_pars, pk_covs)
            ]

            cal_fwhms, cal_fwhms_errs = zip(*cal_fwhms)
            cal_fwhms = np.asarray(cal_fwhms)
            cal_fwhms_errs = np.asarray(cal_fwhms_errs)
            self.results["pk_fwhms"] = np.asarray(
                [(u, e) for u, e in zip(cal_fwhms, cal_fwhms_errs)]
            )

            log.info(f"{len(cal_fwhms)} FWHMs found:")
            log.info(f"\t   Energy   | FWHM  ")
            for i, (Ei, fwhm, fwhme) in enumerate(
                zip(fitted_peaks_keV, cal_fwhms, cal_fwhms_errs)
            ):
                log.info(
                    f"\t{i}".ljust(4)
                    + str(Ei).ljust(9)
                    + f"| {fwhm:.2f}+-{fwhme:.2f} keV".ljust(5)
                )
            self.fit_energy_res()
            log.debug(f"high stats fitting successful")
        except:
            self.results = {}
            log.debug(f"high stats fitting failed")


def get_peak_labels(
    labels: list[str], pars: list[float]
) -> tuple(list[float], list[float]):
    out = []
    out_labels = []
    for i, label in enumerate(labels):
        if i % 2 == 1:
            continue
        else:
            out.append(f"{pgf.poly(label, pars):.1f}")
            out_labels.append(label)
    return out_labels, out


def get_peak_label(peak: float) -> str:
    if peak == 583.191:
        return "Tl 583"
    elif peak == 727.33:
        return "Bi 727"
    elif peak == 860.564:
        return "Tl 860"
    elif peak == 1592.53:
        return "Tl DEP"
    elif peak == 1620.5:
        return "Bi FEP"
    elif peak == 2103.53:
        return "Tl SEP"
    elif peak == 2614.5:
        return "Tl FEP"
    else:
        return ""


def plot_fits(
    ecal_class, data, figsize=[12, 8], fontsize=12, ncols=3, nrows=3, binning_keV=5
):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fitted_peaks = ecal_class.results["got_peaks_keV"]
    pk_pars = ecal_class.results["pk_pars"]
    pk_ranges = ecal_class.results["pk_ranges"]
    p_vals = ecal_class.results["pk_pvals"]

    fitted_gof_funcs = []
    for i, peak in enumerate(ecal_class.glines):
        if peak in fitted_peaks:
            fitted_gof_funcs.append(ecal_class.gof_funcs[i])

    mus = [
        pgf.get_mu_func(func_i, pars_i) if pars_i is not None else np.nan
        for func_i, pars_i in zip(fitted_gof_funcs, pk_pars)
    ]

    fig = plt.figure()
    derco = np.polyder(np.poly1d(ecal_class.pars)).coefficients
    der = [pgf.poly(5, derco) for Ei in fitted_peaks]
    for i, peak in enumerate(mus):
        range_adu = 5 / der[i]
        plt.subplot(nrows, ncols, i + 1)
        try:
            binning = np.arange(pk_ranges[i][0], pk_ranges[i][1], 0.1 / der[i])
            bin_cs = (binning[1:] + binning[:-1]) / 2
            energies = data.query(
                f"{ecal_class.energy_param}>{pk_ranges[i][0]}&{ecal_class.energy_param}<{pk_ranges[i][1]}&{ecal_class.selection_string}"
            )[ecal_class.energy_param]
            energies = energies.iloc[: ecal_class.n_events]

            counts, bs, bars = plt.hist(energies, bins=binning, histtype="step")
            if pk_pars[i] is not None:
                fit_vals = (
                    fitted_gof_funcs[i](bin_cs, *pk_pars[i][:-1], 0) * np.diff(bs)[0]
                )
                plt.plot(bin_cs, fit_vals)
                plt.step(
                    bin_cs,
                    [
                        (fval - count) / count if count != 0 else (fval - count)
                        for count, fval in zip(counts, fit_vals)
                    ],
                    where="mid",
                )

                plt.annotate(
                    get_peak_label(fitted_peaks[i]),
                    (0.02, 0.9),
                    xycoords="axes fraction",
                )
                plt.annotate(
                    f"{fitted_peaks[i]:.1f} keV", (0.02, 0.8), xycoords="axes fraction"
                )
                plt.annotate(
                    f"p-value : {p_vals[i]:.4f}", (0.02, 0.7), xycoords="axes fraction"
                )
                plt.xlabel("Energy (keV)")
                plt.ylabel("Counts")
                plt.legend(loc="upper left", frameon=False)
                plt.xlim([peak - range_adu, peak + range_adu])
                locs, labels = plt.xticks()
                new_locs, new_labels = get_peak_labels(locs, ecal_class.pars)
                plt.xticks(ticks=new_locs, labels=new_labels)
        except:
            pass

    plt.tight_layout()
    plt.close()
    return fig


def plot_2614_timemap(
    ecal_class,
    data,
    figsize=[12, 8],
    fontsize=12,
    erange=[2580, 2630],
    dx=1,
    time_dx=180,
):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    selection = data.query(
        f"{ecal_class.cal_energy_param}>2560&{ecal_class.cal_energy_param}<2660&{ecal_class.selection_string}"
    )

    fig = plt.figure()
    if len(selection) == 0:
        pass
    else:
        time_bins = np.arange(
            (np.amin(data["timestamp"]) // time_dx) * time_dx,
            ((np.amax(data["timestamp"]) // time_dx) + 2) * time_dx,
            time_dx,
        )

        plt.hist2d(
            selection["timestamp"],
            selection[ecal_class.cal_energy_param],
            bins=[time_bins, np.arange(erange[0], erange[1] + dx, dx)],
            norm=LogNorm(),
        )

    ticks, labels = plt.xticks()
    plt.xlabel(
        f"Time starting : {datetime.utcfromtimestamp(ticks[0]).strftime('%d/%m/%y %H:%M')}"
    )
    plt.ylabel("Energy(keV)")
    plt.ylim([erange[0], erange[1]])

    plt.xticks(
        ticks,
        [datetime.utcfromtimestamp(tick).strftime("%H:%M") for tick in ticks],
    )
    plt.close()
    return fig


def plot_pulser_timemap(
    ecal_class,
    data,
    pulser_field="is_pulser",
    figsize=[12, 8],
    fontsize=12,
    dx=0.2,
    time_dx=180,
    n_spread=3,
):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    time_bins = np.arange(
        (np.amin(data["timestamp"]) // time_dx) * time_dx,
        ((np.amax(data["timestamp"]) // time_dx) + 2) * time_dx,
        time_dx,
    )

    selection = data.query(pulser_field)
    fig = plt.figure()
    if len(selection) == 0:
        pass

    else:
        mean = np.nanpercentile(selection[ecal_class.cal_energy_param], 50)
        spread = mean - np.nanpercentile(selection[ecal_class.cal_energy_param], 10)

        plt.hist2d(
            selection["timestamp"],
            selection[ecal_class.cal_energy_param],
            bins=[
                time_bins,
                np.arange(mean - n_spread * spread, mean + n_spread * spread + dx, dx),
            ],
            norm=LogNorm(),
        )
        plt.ylim([mean - n_spread * spread, mean + n_spread * spread])
    ticks, labels = plt.xticks()
    plt.xlabel(
        f"Time starting : {datetime.utcfromtimestamp(ticks[0]).strftime('%d/%m/%y %H:%M')}"
    )
    plt.ylabel("Energy(keV)")

    plt.xticks(
        ticks,
        [datetime.utcfromtimestamp(tick).strftime("%H:%M") for tick in ticks],
    )
    plt.close()
    return fig


def bin_pulser_stability(ecal_class, data, pulser_field="is_pulser", time_slice=180):
    selection = data.query(pulser_field)

    utime_array = data["timestamp"]
    select_energies = selection[ecal_class.cal_energy_param].to_numpy()

    time_bins = np.arange(
        (np.amin(utime_array) // time_slice) * time_slice,
        ((np.amax(utime_array) // time_slice) + 2) * time_slice,
        time_slice,
    )
    # bin time values
    times_average = (time_bins[:-1] + time_bins[1:]) / 2

    if len(selection) == 0:
        return {
            "time": times_average,
            "energy": np.full_like(times_average, np.nan),
            "spread": np.full_like(times_average, np.nan),
        }

    nanmedian = (
        lambda x: np.nanpercentile(x, 50) if len(x[~np.isnan(x)]) >= 10 else np.nan
    )
    error = (
        lambda x: np.nanvar(x) / np.sqrt(len(x))
        if len(x[~np.isnan(x)]) >= 10
        else np.nan
    )

    par_average, _, _ = binned_statistic(
        selection["timestamp"], select_energies, statistic=nanmedian, bins=time_bins
    )
    par_error, _, _ = binned_statistic(
        selection["timestamp"], select_energies, statistic=error, bins=time_bins
    )

    return {"time": times_average, "energy": par_average, "spread": par_error}


def bin_stability(ecal_class, data, time_slice=180, energy_range=[2585, 2660]):
    selection = data.query(
        f"{ecal_class.cal_energy_param}>{energy_range[0]}&{ecal_class.cal_energy_param}<{energy_range[1]}&{ecal_class.selection_string}"
    )

    utime_array = data["timestamp"]
    select_energies = selection[ecal_class.cal_energy_param].to_numpy()

    time_bins = np.arange(
        (np.amin(utime_array) // time_slice) * time_slice,
        ((np.amax(utime_array) // time_slice) + 2) * time_slice,
        time_slice,
    )
    # bin time values
    times_average = (time_bins[:-1] + time_bins[1:]) / 2

    if len(selection) == 0:
        return {
            "time": times_average,
            "energy": np.full_like(times_average, np.nan),
            "spread": np.full_like(times_average, np.nan),
        }

    nanmedian = (
        lambda x: np.nanpercentile(x, 50) if len(x[~np.isnan(x)]) >= 10 else np.nan
    )
    error = (
        lambda x: np.nanvar(x) / np.sqrt(len(x))
        if len(x[~np.isnan(x)]) >= 10
        else np.nan
    )

    par_average, _, _ = binned_statistic(
        selection["timestamp"], select_energies, statistic=nanmedian, bins=time_bins
    )
    par_error, _, _ = binned_statistic(
        selection["timestamp"], select_energies, statistic=error, bins=time_bins
    )

    return {"time": times_average, "energy": par_average, "spread": par_error}


def plot_cal_fit(ecal_class, data, figsize=[12, 8], fontsize=12, erange=[200, 2700]):
    pk_pars = ecal_class.results["pk_pars"]
    fitted_peaks = ecal_class.results["got_peaks_keV"]
    pk_errs = ecal_class.results["pk_errors"]

    fitted_gof_funcs = []
    for i, peak in enumerate(ecal_class.glines):
        if peak in fitted_peaks:
            fitted_gof_funcs.append(ecal_class.gof_funcs[i])

    mus = [
        pgf.get_mu_func(func_i, pars_i) if pars_i is not None else np.nan
        for func_i, pars_i in zip(fitted_gof_funcs, pk_pars)
    ]

    mu_errs = [
        pgf.get_mu_func(func_i, pars_i) if pars_i is not None else np.nan
        for func_i, pars_i in zip(fitted_gof_funcs, pk_errs)
    ]

    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig, (ax1, ax2) = plt.subplots(
        2, 1, sharex=True, gridspec_kw={"height_ratios": [3, 1]}
    )

    cal_bins = np.linspace(0, np.nanmax(mus) * 1.1, 20)

    ax1.scatter(fitted_peaks, mus, marker="x", c="b")

    ax1.plot(pgf.poly(cal_bins, ecal_class.pars), cal_bins, lw=1, c="g")

    ax1.grid()
    ax1.set_xlim([erange[0], erange[1]])
    ax1.set_ylabel("Energy (ADC)")
    ax2.errorbar(
        fitted_peaks,
        pgf.poly(np.array(mus), ecal_class.pars) - fitted_peaks,
        yerr=pgf.poly(np.array(mus) + np.array(mu_errs), ecal_class.pars)
        - pgf.poly(np.array(mus), ecal_class.pars),
        linestyle=" ",
        marker="x",
        c="b",
    )
    ax2.grid()
    ax2.set_xlabel("Energy (keV)")
    ax2.set_ylabel("Residuals (keV)")
    plt.close()
    return fig


def plot_eres_fit(ecal_class, data, erange=[200, 2700], figsize=[12, 8], fontsize=12):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fwhms = ecal_class.results["pk_fwhms"][:, 0]
    dfwhms = ecal_class.results["pk_fwhms"][:, 1]
    fitted_peaks = ecal_class.results["fitted_keV"]

    #####
    # Remove the Tl SEP and DEP from calibration if found
    fwhm_peaks = np.array([], dtype=np.float32)
    indexes = []
    for i, peak in enumerate(fitted_peaks):
        if peak == 2103.53:
            log.info(f"Tl SEP found at index {i}")
            indexes.append(i)
            continue
        elif peak == 1592.53:
            log.info(f"Tl DEP found at index {i}")
            indexes.append(i)
            continue
        elif np.isnan(dfwhms[i]):
            log.info(f"{peak} failed")
            indexes.append(i)
            continue
        elif peak == 511.0:
            log.info(f"e annhilation found at index {i}")
            indexes.append(i)
            continue
        else:
            fwhm_peaks = np.append(fwhm_peaks, peak)
    fit_fwhms = np.delete(fwhms, [indexes])
    fit_dfwhms = np.delete(dfwhms, [indexes])

    fig, (ax1, ax2) = plt.subplots(
        2, 1, sharex=True, gridspec_kw={"height_ratios": [3, 1]}
    )
    ax1.errorbar(fwhm_peaks, fit_fwhms, yerr=fit_dfwhms, marker="x", lw=0, c="black")

    fwhm_slope_bins = np.arange(erange[0], erange[1], 10)

    qbb_line_vx = [2039.0, 2039.0]
    qbb_line_vy = [
        0.9
        * np.nanmin(
            fwhm_linear.func(fwhm_slope_bins, *ecal_class.fwhm_fit_linear["parameters"])
        ),
        np.nanmax(
            [
                ecal_class.fwhm_fit_linear["Qbb_fwhm_in_keV"],
                ecal_class.fwhm_fit_quadratic["Qbb_fwhm_in_keV"],
            ]
        ),
    ]
    qbb_line_hx = [erange[0], 2039.0]

    ax1.plot(
        fwhm_slope_bins,
        fwhm_linear.func(fwhm_slope_bins, *ecal_class.fwhm_fit_linear["parameters"]),
        lw=1,
        c="g",
        label=f'linear, Qbb fwhm: {ecal_class.fwhm_fit_linear["Qbb_fwhm_in_keV"]:1.2f} +- {ecal_class.fwhm_fit_linear["Qbb_fwhm_err_in_keV"]:1.2f} keV',
    )
    ax1.plot(
        fwhm_slope_bins,
        fwhm_quadratic.func(
            fwhm_slope_bins, *ecal_class.fwhm_fit_quadratic["parameters"]
        ),
        lw=1,
        c="b",
        label=f'quadratic, Qbb fwhm: {ecal_class.fwhm_fit_quadratic["Qbb_fwhm_in_keV"]:1.2f} +- {ecal_class.fwhm_fit_quadratic["Qbb_fwhm_err_in_keV"]:1.2f} keV',
    )
    ax1.plot(
        qbb_line_hx,
        [
            ecal_class.fwhm_fit_linear["Qbb_fwhm_in_keV"],
            ecal_class.fwhm_fit_linear["Qbb_fwhm_in_keV"],
        ],
        lw=1,
        c="r",
        ls="--",
    )
    ax1.plot(
        qbb_line_hx,
        [
            ecal_class.fwhm_fit_quadratic["Qbb_fwhm_in_keV"],
            ecal_class.fwhm_fit_quadratic["Qbb_fwhm_in_keV"],
        ],
        lw=1,
        c="r",
        ls="--",
    )
    ax1.plot(qbb_line_vx, qbb_line_vy, lw=1, c="r", ls="--")

    ax1.legend(loc="upper left", frameon=False)
    if np.isnan(ecal_class.fwhm_fit_linear["parameters"]).all():
        [
            0.9 * np.nanmin(fit_fwhms),
            1.1 * np.nanmax(fit_fwhms),
        ]
    else:
        ax1.set_ylim(
            [
                0.9
                * np.nanmin(
                    fwhm_linear.func(
                        fwhm_slope_bins, *ecal_class.fwhm_fit_linear["parameters"]
                    )
                ),
                1.1
                * np.nanmax(
                    fwhm_linear.func(
                        fwhm_slope_bins, *ecal_class.fwhm_fit_linear["parameters"]
                    )
                ),
            ]
        )
    ax1.set_xlim(erange)
    ax1.set_ylabel("FWHM energy resolution (keV)")
    ax2.plot(
        fwhm_peaks,
        (
            fit_fwhms
            - fwhm_linear.func(fwhm_peaks, *ecal_class.fwhm_fit_linear["parameters"])
        )
        / fit_dfwhms,
        lw=0,
        marker="x",
        c="g",
    )
    ax2.plot(
        fwhm_peaks,
        (
            fit_fwhms
            - fwhm_quadratic.func(
                fwhm_peaks, *ecal_class.fwhm_fit_quadratic["parameters"]
            )
        )
        / fit_dfwhms,
        lw=0,
        marker="x",
        c="b",
    )
    ax2.plot(erange, [0, 0], color="black", lw=0.5)
    ax2.set_xlabel("Energy (keV)")
    ax2.set_ylabel("Normalised Residuals")
    plt.tight_layout()
    plt.close()
    return fig


def bin_spectrum(
    ecal_class,
    data,
    cut_field="is_valid_cal",
    pulser_field="is_pulser",
    erange=[0, 3000],
    dx=2,
):
    bins = np.arange(erange[0], erange[1] + dx, dx)
    return {
        "bins": pgh.get_bin_centers(bins),
        "counts": np.histogram(
            data.query(ecal_class.selection_string)[ecal_class.cal_energy_param], bins
        )[0],
        "cut_counts": np.histogram(
            data.query(f"(~{cut_field})&(~{pulser_field})")[
                ecal_class.cal_energy_param
            ],
            bins,
        )[0],
        "pulser_counts": np.histogram(
            data.query(pulser_field)[ecal_class.cal_energy_param],
            bins,
        )[0],
    }


def bin_survival_fraction(
    ecal_class,
    data,
    cut_field="is_valid_cal",
    pulser_field="is_pulser",
    erange=[0, 3000],
    dx=6,
):
    counts_pass, bins_pass, _ = pgh.get_hist(
        data.query(ecal_class.selection_string)[ecal_class.cal_energy_param],
        bins=np.arange(erange[0], erange[1] + dx, dx),
    )
    counts_fail, bins_fail, _ = pgh.get_hist(
        data.query(f"(~{cut_field})&(~{pulser_field})")[ecal_class.cal_energy_param],
        bins=np.arange(erange[0], erange[1] + dx, dx),
    )
    sf = 100 * (counts_pass + 10 ** (-6)) / (counts_pass + counts_fail + 10 ** (-6))
    return {"bins": pgh.get_bin_centers(bins_pass), "sf": sf}
